library(shipunov)

## function to calculate number of taxonomically informative characters
Is.tax.inform.char <- function(vec) sum(table(vec, useNA="no") > 1) > 1

fragments <- c("its", "rbcl", "trnl", "matk", "petd") # fragile
treesp <- read.table("../../data/buxineae_treesp.txt", sep="\t", h=T, as.is=T)
sequences <- read.table("../../data/buxineae_dna.txt", sep="\t", h=T, as.is=T, quote="")
outgroups <- treesp$SPECIES.NEW[treesp$TYPE == "outgroup"]

dna <- structure(vector("list", 5), names=fragments)
for (f in fragments) {
dna[[f]] <- Read.fasta(paste0("32_alignments_trimmed_gapcoded/", f, ".fasta"))
sp.id <- do.call(rbind, strsplit(dna[[f]]$name, "__"))
dna[[f]] <- data.frame(SP=sp.id[, 1], ID=sp.id[, 2], SEQ=dna[[f]][, 2], stringsAsFactors=F)
## dna[[f]] <- dna[[f]][order(nchar(gsub("-", "", dna[[f]]$SEQ)), decreasing=TRUE), ]
## dna[[f]] <- dna[[f]][order(dna[[f]]$SP), ]
dna[[f]] <- dna[[f]][order(dna[[f]]$SP, nchar(gsub("-", "", dna[[f]]$SEQ))), ]
}

## join all sp/ids from list
sp.ids <- do.call(rbind, c(lapply(dna, "[", 1:2), make.row.names=F))
tmp1 <- suppressWarnings(do.call(rbind, strsplit(sp.ids$ID, ", ")))
tmp2 <- unique(sort(paste(c(tmp1), sp.ids$SP, sep="__")))
ids.sp <- do.call(rbind, strsplit(tmp2, "__"))

## ===
## ===
## ===

## STEP 1. STRICT: its, rbcl and trnl merged on "our" ids

## rows
tmp1 <- sequences[sequences$SOURCE == "MISU", "SEQUENCE.ID"] # not in buxineae (more uniform ids there)
tmp2 <- suppressWarnings(do.call(rbind, strsplit(tmp1, ", ")))
ourids.all <- unique(sort(tmp2))
# ourids <- ids.sp[ids.sp[, 1] %in% ourids.all, ] # defunct at the moment
ourids <- ids.sp[grepl("B-|slx", ids.sp[, 1]), ] # fragile
cat("Check duplicated 'our' ids:\n") # do not remember why, but let it stay
ourids[, 1][duplicated(ourids)]

## make
strict <- data.frame(ourids, stringsAsFactors=F)
names(strict) <- c("ID", "SP")

## outgroups first, other sorted
tmp1 <- strict[strict$SP %in% outgroups, ]
tmp2 <- strict[!strict$SP %in% outgroups, ]
tmp2 <- tmp2[order(tmp2$SP), ]
strict <- rbind(tmp1, tmp2)

for (n in fragments[1:3]) { # fragile
 strict[[n]] <- NA
 for (i in 1:nrow(strict)) {
  ii <- dna[[n]]$SEQ[grepl(strict$ID[i], dna[[n]]$ID)]
  strict[[n]][i] <- ifelse(length(ii) == 0, NA, ii)
 }
}

## ===
## ===
## ===

## STEP 2. SEMI-STRICT (the main matrix):
## (1) add petd, matk and rbcl; 
## (2) add new species which were not in "our" ids;
## (3) fill species sequences into all NA gaps and keep that id

semistrict <- strict

## rows
ssp <- sort(unique(semistrict$SP))
allsp <- sort(unique(sp.ids$SP))
addsp <- allsp[!allsp %in% ssp]
semistrict[(nrow(semistrict) + 1) : (nrow(semistrict) + length(addsp)), ] <- NA
semistrict[nrow(semistrict) : (nrow(semistrict) - (length(addsp) - 1)), "SP"] <- addsp
semistrict[nrow(semistrict) : (nrow(semistrict) - (length(addsp) - 1)), "ID"] <- NA

## outgroups first, other sorted
tmp1 <- semistrict[semistrict$SP %in% outgroups, ]
tmp2 <- semistrict[!semistrict$SP %in% outgroups, ]
tmp2 <- tmp2[order(tmp2$SP), ]
semistrict <- rbind(tmp1, tmp2)

## columns
names.tmp3 <- c(paste0(fragments, ".ID"), fragments[4:5]) # fragile
tmp3 <- structure(data.frame(matrix(NA, nrow=nrow(semistrict), ncol=length(names.tmp3))), names=names.tmp3)
semistrict <- cbind(semistrict, tmp3)
semistrict <- semistrict[, c("SP", "ID", paste0(fragments, ".ID"), fragments)]

## main
for(n in fragments) {
 d <- dna[[n]]
 f.id <- paste0(n, ".ID")
 for (i in 1:nrow(semistrict)) {
  if (is.na(semistrict[i, n])) {
  dd <- d$SEQ[grepl(semistrict$SP[i], d$SP)]
  ii <- d$ID[grepl(semistrict$SP[i], d$SP)]
  if (length(dd) > 0 & length(ii) > 0) {
   semistrict[i, n] <- dd[1]
   semistrict[i, f.id] <- ii[1]
   } else {
   semistrict[i, f.id] <- "-"
   }
 }}
}

## STEP 3. Write files
## (1) semistrict (main) => will be used in kmer tree
## (2) our (its + trnl + rbcl) => will be used in oi
## (3) cp (trnl + rbcl + petd + matk)
## (4) oi (imputed) => in the separate file

## semistrict statistics
semistrict.m <- data.frame(semistrict[, 1:2], apply(semistrict[, fragments], 2, function(.y) ifelse(is.na(.y), 0, 1)))
semistrict.m$COUNT <- rowSums(semistrict.m[, fragments])
## what missed after concatenation
semistrict.m[order(semistrict.m$COUNT, decreasing=T), ]
## how many ids in each missing state
table(apply(semistrict[, fragments], 1, function(.y) sum(!is.na(.y))))
## how many ids missed in each fragment
apply(semistrict[, fragments], 2, function(.x) paste0(sum(is.na(.x)), "/", length(.x)))
## how many sequences our and total
all.ids <- unlist(semistrict[, c("ID", paste0(fragments, ".ID"))])
## our
sum(all.ids %in% ourids.all & !is.na(all.ids) & all.ids != "-")
## total
sum(!is.na(all.ids) & all.ids != "-")
## number of potentially informative characters and total number of characters per fragment
for (f in fragments) {
cat(f)
tmp <- semistrict[, f]
tmp <- tmp[!is.na(tmp)]
tmp.table <- do.call(rbind, strsplit(tmp, split=""))
tmp.table[tmp.table == "-" | tmp.table == "N"] <- NA
cat("\tall chars ", ncol(tmp.table))
tmp.itic <- apply(tmp.table, 2, Is.tax.inform.char)
cat("\tpotentially taxonomically informative chars ", sum(tmp.itic), "\n")
}

## write original semistrict matrix
write.table(semistrict, file=paste0("40_concatenated/", "semistrict_orig.txt"), sep="\t", row.names=F, quote=F)

### write all duplicates, no gaps and N's FASTA (=> kmer)
seq.tmp <- apply(semistrict[, fragments], 2, function(.x) gsub("-", "", .x))
seq.tmp <- apply(seq.tmp, 2, function(.x) ifelse(is.na(.x), "", .x))
seq.tmp <- apply(seq.tmp, 1, function(.x) paste(.x, collapse=""))
## make joint ids
kmer.tmp <- semistrict
kmer.tmp.f.ids <- paste0(fragments, ".ID")
ids.tmp1 <- apply(kmer.tmp[, c("SP", "ID")], 1, function(.x) paste(.x, collapse="__"))
for (i in kmer.tmp.f.ids) kmer.tmp[, i] <- ifelse(is.na(kmer.tmp[, i]), "", paste(sub(".ID", "", i), kmer.tmp[, i]))
ids.tmp2 <- apply(kmer.tmp[, kmer.tmp.f.ids], 1, function(.x) paste(.x, collapse=" "))
ids.tmp2 <- gsub("(^ +)|( +$)", "", ids.tmp2)
ids.tmp2 <- gsub("( +)", " ", ids.tmp2)
ids.tmp <- paste(ids.tmp1, ids.tmp2)
ids.tmp <- gsub(" +", "_", ids.tmp)
ids.tmp <- gsub(",", "", ids.tmp)
ids.tmp <- gsub("_$", "", ids.tmp)
ids.tmp <- gsub("_NA_", "_", ids.tmp)
##
kmer <- data.frame(ID=ids.tmp, SEQ=seq.tmp, stringsAsFactors=F)
Write.fasta(kmer, file=paste0("40_concatenated/", "kmer.fasta"))

## ===
## make "our" its/rbcl/trnl (will go to its imputation, and to all trees)
## (1) subset
## (2) make joint ids
## (3) remove duplicates and make combined ids
## (4) NA's into N's
## (5) write FASTA
## ===

our.f <- c("its", "rbcl", "trnl")
our.f.ids <- paste0(our.f, ".ID")
our.tmp <- semistrict[, c("SP", "ID", our.f.ids, our.f)]
our.tmp <- our.tmp[!rowSums(sapply(our.tmp[, our.f], is.na)) == length(our.f), ] # remove rows where all fragmnents missed

## make joint ids
ids.tmp0 <- our.tmp[, c("SP", "ID", our.f.ids)]
ids.tmp1 <- apply(ids.tmp0[, c("SP", "ID")], 1, function(.x) paste(.x, collapse="__"))
for (i in our.f.ids) ids.tmp0[, i] <- ifelse(is.na(ids.tmp0[, i]), "", paste(sub(".ID", "", i), ids.tmp0[, i]))
ids.tmp2 <- apply(ids.tmp0[, our.f.ids], 1, function(.x) paste(.x, collapse=" "))
ids.tmp2 <- gsub("(^ +)|( +$)", "", ids.tmp2)
ids.tmp2 <- gsub("( +)", " ", ids.tmp2)
ids.tmp <- paste(ids.tmp1, ids.tmp2)
ids.tmp <- gsub(" +", "_", ids.tmp)
ids.tmp <- gsub(",", "", ids.tmp)
ids.tmp <- gsub("_$", "", ids.tmp)
ids.tmp <- gsub("_NA_", "_", ids.tmp)

## duplicates
groups <- as.numeric(as.factor(apply(our.tmp[, c("SP", our.f)], 1, paste, collapse="")))
comb.ids <-  aggregate(ids.tmp, by=list(groups), paste, collapse=" + ")
our.tmp <- our.tmp[!duplicated(groups), ] # remove duplicates in main table
ids <- ids.tmp[!duplicated(groups)] # remove duplicates in joint ids
comb.ids <- comb.ids[groups[!duplicated(groups)], 2] # the same order as in deduplicated data
## this table is a "filter" for tree labels
write.table(data.frame(ids, comb.ids), file="40_concatenated/our_comb_ids.txt", sep="\t", quote=F, row.names=F)

## write our matrix (=> oi)
write.table(our.tmp, file=paste0("40_concatenated/", "our.txt"), sep="\t", row.names=F, quote=F)

## our.tmp statistics
our.tmp.m <- data.frame(our.tmp[, 1:2], apply(our.tmp[, our.f], 2, function(.y) ifelse(is.na(.y), 0, 1)))
our.tmp.m$COUNT <- rowSums(our.tmp.m[, our.f])
## what missed after concatenation
our.tmp.m[order(our.tmp.m$COUNT, decreasing=T), ]
## how many ids in each missing state
table(apply(our.tmp[, our.f], 1, function(.y) sum(!is.na(.y))))
## how many ids missed in each fragment
apply(our.tmp[, our.f], 2, function(.x) paste0(sum(is.na(.x)), "/", length(.x)))
## how many sequences our and total
all.ids <- unlist(our.tmp[, c("ID", paste0(our.f, ".ID"))])
## our
sum(all.ids %in% ourids.all & !is.na(all.ids) & all.ids != "-")
## total
sum(!is.na(all.ids) & all.ids != "-")
## number of potentially informative characters and total number of characters per fragment
for (f in our.f) {
cat(f)
tmp <- our.tmp[, f]
tmp <- tmp[!is.na(tmp)]
tmp.table <- do.call(rbind, strsplit(tmp, split=""))
tmp.table[tmp.table == "-" | tmp.table == "N"] <- NA
cat("\tall chars ", ncol(tmp.table))
tmp.itic <- apply(tmp.table, 2, Is.tax.inform.char)
cat("\tpotentially taxonomically informative chars ", sum(tmp.itic), "\n")
}

## replace NA's with N's and write resulted FASTA
for (i in our.f) our.tmp[, i] <- ifelse(is.na(our.tmp[, i]), paste(rep("N", max(nchar(our.tmp[, i]), na.rm=T)), collapse=""), our.tmp[, i])
seq <- apply(our.tmp[, our.f], 1, function(.x) paste(.x, collapse=""))
our.tmp.new <- data.frame(ID=ids, SEQ=seq, stringsAsFactors=F)
Write.fasta(our.tmp.new, file=paste0("40_concatenated/", "our.fasta"))

## ===
## make "cp" rbcl/trnl/matk/petd and  (will go to all trees)
## [generalized into "data" for semistrict subset and "data.f" for fragment list,
## only (a) file names, (b) comments and (c) removal missed markers row need attention]

data.f <- c("rbcl", "trnl", "matk", "petd")
data.f.ids <- paste0(data.f, ".ID")
data <- semistrict[, c("SP", "ID", data.f.ids, data.f)]
data <- data[!rowSums(sapply(data[, data.f], is.na)) > 1, ] # keep rows only with 3 non-missed fragments

## make joint ids
ids.tmp0 <- data[, c("SP", "ID", data.f.ids)]
ids.tmp1 <- apply(ids.tmp0[, c("SP", "ID")], 1, function(.x) paste(.x, collapse="__"))
for (i in data.f.ids) ids.tmp0[, i] <- ifelse(is.na(ids.tmp0[, i]), "", paste(sub(".ID", "", i), ids.tmp0[, i]))
ids.tmp2 <- apply(ids.tmp0[, data.f.ids], 1, function(.x) paste(.x, collapse=" "))
ids.tmp2 <- gsub("(^ +)|( +$)", "", ids.tmp2)
ids.tmp2 <- gsub("( +)", " ", ids.tmp2)
ids.tmp <- paste(ids.tmp1, ids.tmp2)
ids.tmp <- gsub(" +", "_", ids.tmp)
ids.tmp <- gsub(",", "", ids.tmp)
ids.tmp <- gsub("_$", "", ids.tmp)
ids.tmp <- gsub("_NA_", "_", ids.tmp)

## duplicates
groups <- as.numeric(as.factor(apply(data[, c("SP", data.f)], 1, paste, collapse="")))
comb.ids <-  aggregate(ids.tmp, by=list(groups), paste, collapse=" + ")
data <- data[!duplicated(groups), ] # remove duplicates in main table
ids <- ids.tmp[!duplicated(groups)] # remove duplicates in joint ids
comb.ids <- comb.ids[groups[!duplicated(groups)], 2] # the same order as in deduplicated data
## this table is a "filter" for tree labels
write.table(data.frame(ids, comb.ids), file="40_concatenated/cp_comb_ids.txt", sep="\t", quote=F, row.names=F)

## cp statistics
data.m <- data.frame(data[, c("SP", "ID")], apply(data[, data.f], 2, function(.y) ifelse(is.na(.y), 0, 1)))
data.m$COUNT <- rowSums(data.m[, data.f])
## what missed after concatenation
data.m[order(data.m$COUNT, decreasing=T), ]
## how many ids in each missing state
table(apply(data[, data.f], 1, function(.y) sum(!is.na(.y))))
## how many ids missed in each fragment
apply(data[, data.f], 2, function(.x) paste0(sum(is.na(.x)), "/", length(.x)))
## how many sequences our and total
all.ids <- unlist(data[, c("ID", paste0(data.f, ".ID"))])
## our
sum(all.ids %in% ourids.all & !is.na(all.ids) & all.ids != "-")
## total
sum(!is.na(all.ids) & all.ids != "-")
## number of potentially informative characters and total number of characters per fragment
for (f in data.f) {
cat(f)
tmp <- data[, f]
tmp <- tmp[!is.na(tmp)]
tmp.table <- do.call(rbind, strsplit(tmp, split=""))
tmp.table[tmp.table == "-" | tmp.table == "N"] <- NA
cat("\tall chars ", ncol(tmp.table))
tmp.itic <- apply(tmp.table, 2, Is.tax.inform.char)
cat("\tpotentially taxonomically informative chars ", sum(tmp.itic), "\n")
}

## replace NA's with N's and write resulted FASTA
for (i in data.f) data[, i] <- ifelse(is.na(data[, i]), paste(rep("N", max(nchar(data[, i]), na.rm=T)), collapse=""), data[, i])
seq <- apply(data[, data.f], 1, function(.x) paste(.x, collapse=""))
data.new <- data.frame(ID=ids, SEQ=seq, stringsAsFactors=F)
Write.fasta(data.new, file=paste0("40_concatenated/", "cp.fasta"))
